import cv2
import numpy as np
from tqdm import tqdm
from collections import defaultdict
import os

GOAL_COLOR = (234, 255, 0)
PLAYER_1_COLOR = (74, 214, 39)
PLAYER_1_AREAL_COLOR = (173, 185, 58)
PLAYER_2_COLOR = (210, 219, 105)
PLAYER_2_AREAL_COLOR = (241, 184, 53)
PLAYER_3_COLOR = (46, 171, 255)
PLAYER_3_AREAL_COLOR = (79, 78, 127)
PLAYER_4_COLOR = (255, 107, 107)
PLAYER_4_AREAL_COLOR = (239, 59, 74)
TAILE_BACKGROUND_COLOR = (174, 95, 74)
BOARD_BACKGROUND_COLOR = (186, 202, 215)

TAILE_SICE_FULL = np.array([64, 64])
TAILE_PIECE_R = 20
BOARD_TAILE_SIZE = np.array([17, 20])
BOARS_SIZE = TAILE_SICE_FULL * BOARD_TAILE_SIZE

COLOR_TEST_TAILES = [
    [[8, 3], [8, 4], [8, 5], [8, 6]],
    [[8, 10], [8, 11], [8, 12], [8, 13]],
    [[6, 8], [5, 8], [4, 8], [3, 8]],
    [[10, 8], [11, 8], [12, 8], [13, 8]],
    [[7, 7], [7, 9], [9, 7], [9, 9]],
    [[7, 6], [7, 10], [9, 6], [9, 10]],
    [[13, 4], [4, 3], [3, 12], [12, 13]]
]

GOAL_TAILES = [
    [7, 7],
    [7, 8],
    [7, 9],
    [8, 7],
    [8, 8],
    [8, 9],
    [9, 7],
    [9, 8],
    [9, 9]
]

PLAYER_1_GOAL_AREAL = [
    [14, 7],
    [14, 8],
    [13, 8],
    [12, 8],
    [11, 8],
    [10, 8],
]
PLAYER_1_HOME_TAILES = [
    [12, 3],
    [14, 3],
    [12, 5],
    [14, 5]
]

PLAYER_2_GOAL_AREAL = [
    [7, 2],
    [8, 2],
    [8, 3],
    [8, 4],
    [8, 5],
    [8, 6]
]

PLAYER_2_HOME_TAILES = [
    [5, 2],
    [5, 4],
    [3, 4],
    [3, 2]
]

PLAYER_3_GOAL_AREAL = [
    [2, 9],
    [2, 8],
    [3, 8],
    [4, 8],
    [5, 8],
    [6, 8]
]

PLAYER_3_HOME_TAILES = [
    [2, 11],
    [2, 13],
    [4, 11],
    [4, 13]
]

PLAYER_4_GOAL_AREAL = [
    [9, 14],
    [8, 14],
    [8, 13],
    [8, 12],
    [8, 11],
    [8, 10]
]

PLAYER_4_HOME_TAILES = [
    [11, 14],
    [13, 14],
    [13, 12],
    [11, 12]
]

STARS_TAILES = [
    [15, 8],
    [10, 7],
    [8, 1],
    [7, 6],
    [1, 8],
    [6, 9],
    [8, 15],
    [9, 10]
]

GLOBS_TIALES = [
    [13, 9],
    [14, 7],
    [9, 3],
    [7, 2],
    [3, 7],
    [2, 9],
    [7, 13],
    [9, 14]
]
FREE_TAIALES = [
    [10, 9],
    [11, 9],
    [12, 9],
    [14, 9],
    [15, 9],
    [15, 7],
    [13, 7],
    [12, 7],
    [11, 7],
    [9, 6],
    [9, 5],
    [9, 4],
    [9, 2],
    [9, 1],
    [7, 1],
    [7, 3],
    [7, 4],
    [7, 5],
    [6, 7],
    [5, 7],
    [4, 7],
    [2, 7],
    [1, 7],
    [1, 9],
    [3, 9],
    [4, 9],
    [5, 9],
    [7, 10],
    [7, 11],
    [7, 12],
    [7, 14],
    [7, 15],
    [9, 15],
    [9, 13],
    [9, 12],
    [9, 11]
]

TRACK_PLAYER_1 = np.array(
    [[14, 7], [13, 7], [12, 7], [11, 7], [10, 7], [9, 6], [9, 5], [9, 4], [9, 3], [9, 2], [9, 1], [8, 1], [7, 1],
     [7, 2], [7, 3], [7, 4], [7, 5], [7, 6], [6, 7], [5, 7], [4, 7], [3, 7], [2, 7], [1, 7], [1, 8], [1, 9], [2, 9],
     [3, 9], [4, 9], [5, 9], [6, 9], [7, 10], [7, 11], [7, 12], [7, 13], [7, 14], [7, 15], [8, 15], [9, 15], [9, 14],
     [9, 13], [9, 12], [9, 11], [9, 10], [10, 9], [11, 9], [12, 9], [13, 9], [14, 9], [15, 9], [15, 8], [15, 7],
     [14, 7], [14, 8], [13, 8], [12, 8], [11, 8], [10, 8], [9, 8]])
DIFF_1 = TRACK_PLAYER_1 - TRACK_PLAYER_1[0]
DIFF_2 = np.array([DIFF_1[:, 1], -DIFF_1[:, 0]]).T
DIFF_3 = np.array([-DIFF_1[:, 0], -DIFF_1[:, 1]]).T
DIFF_4 = np.array([-DIFF_1[:, 1], DIFF_1[:, 0]]).T

TRACK_PLAYER_2 = np.array([7, 2]) + DIFF_2
TRACK_PLAYER_3 = np.array([2, 9]) + DIFF_3
TRACK_PLAYER_4 = np.array([9, 14]) + DIFF_4

HOME_TAILES = {
    0: PLAYER_1_HOME_TAILES,
    1: PLAYER_2_HOME_TAILES,
    2: PLAYER_3_HOME_TAILES,
    3: PLAYER_4_HOME_TAILES
}

TRACKS = {
    0: TRACK_PLAYER_1,
    1: TRACK_PLAYER_2,
    2: TRACK_PLAYER_3,
    3: TRACK_PLAYER_4
}

PLAYER_COLORS = {
    0: PLAYER_1_COLOR,
    1: PLAYER_2_COLOR,
    2: PLAYER_3_COLOR,
    3: PLAYER_4_COLOR
}
folder, _ = os.path.split(__file__)

big_glob = cv2.imread(os.path.join(folder, "glob.png"))
big_glob = cv2.cvtColor(big_glob, cv2.COLOR_BGR2RGB)
small_glob = cv2.resize(big_glob, (40, 40))
glob_mask = cv2.inRange(small_glob, (255, 255, 255), (255, 255, 255)) == 0

big_star = cv2.imread(os.path.join(folder, "star.png"))
big_star = cv2.cvtColor(big_star, cv2.COLOR_BGR2RGB)
small_star = cv2.resize(big_star, (40, 40))
star_mask = cv2.inRange(small_star, (255, 255, 255), (255, 255, 255)) == 0


def get_taile_cord(n, m):
    top_left = (m * TAILE_SICE_FULL[0], n * TAILE_SICE_FULL[1])
    top_right = ((m + 1) * TAILE_SICE_FULL[0], n * TAILE_SICE_FULL[1])
    bot_left = (m * TAILE_SICE_FULL[0], (n + 1) * TAILE_SICE_FULL[1])
    bot_right = ((m + 1) * TAILE_SICE_FULL[0], (n + 1) * TAILE_SICE_FULL[1])
    center = (top_left[0] + int(np.round((top_right[0] - top_left[0]) / 2)),
              top_left[1] + int(np.round((bot_left[1] - top_left[1]) / 2)))

    return top_left, bot_left, top_right, bot_right, center


def draw_tail(img, n, m, line_color=None, fill_color=None, thickness=2):
    top_left, bot_left, top_right, bot_right, center = get_taile_cord(n, m)

    if fill_color is not None:
        cv2.rectangle(img, top_left, bot_right, fill_color, thickness=-1)
    if line_color is not None:
        cv2.rectangle(img, top_left, bot_right, line_color, thickness)


def get_all_tailes_within(n_start, n_end, m_start, m_end):
    return np.concatenate(np.mgrid[n_start:n_end, m_start:m_end].transpose())


def draw_text(board, text, center, color, thickness=1, fontScale=0.5):
    font = cv2.FONT_HERSHEY_SIMPLEX

    (label_width, label_height), baseline = cv2.getTextSize(text, font, fontScale, thickness)

    bot_left = (center[0] - label_width / 2, center[1] + label_height / 2)
    bot_left = (int(np.round(bot_left[0])), int(np.round(bot_left[1])))

    cv2.putText(board, text, bot_left, font, fontScale, color, thickness, cv2.LINE_AA)


def draw_taile_indxs(board):
    for n in range(0, BOARD_TAILE_SIZE[0]):
        for m in range(0, BOARD_TAILE_SIZE[1]):
            top_left, bot_left, top_right, bot_right, center = get_taile_cord(n, m)
            draw_text(board, f"{n},{m}", center, (0, 0, 0))


def draw_piece(board, n, m, amount, color, thickness=5, lineType=8, shift=0):
    top_left, bot_left, top_right, bot_right, center = get_taile_cord(n, m)

    cv2.circle(board, center, TAILE_PIECE_R, color, thickness=thickness, lineType=8, shift=0)
    draw_text(board, f"{amount}", center, color, thickness=2, fontScale=1)


def draw_basic_board(draw_taile_number=False):
    board = np.full(shape=(*BOARS_SIZE, 3), fill_value=BOARD_BACKGROUND_COLOR, dtype=np.uint8)

    draw_multi_box(board, (7, 7), (9, 9), line_color=(0, 0, 0), fill_color=GOAL_COLOR)

    for n, m in STARS_TAILES + GLOBS_TIALES + FREE_TAIALES:
        draw_tail(board, n, m, line_color=(0, 0, 0), fill_color=TAILE_BACKGROUND_COLOR)

    for n, m in PLAYER_1_HOME_TAILES + PLAYER_1_GOAL_AREAL:
        draw_tail(board, n, m, line_color=(0, 0, 0), fill_color=PLAYER_1_AREAL_COLOR)

    for n, m in PLAYER_2_HOME_TAILES + PLAYER_2_GOAL_AREAL:
        draw_tail(board, n, m, line_color=(0, 0, 0), fill_color=PLAYER_2_AREAL_COLOR)

    for n, m in PLAYER_3_HOME_TAILES + PLAYER_3_GOAL_AREAL:
        draw_tail(board, n, m, line_color=(0, 0, 0), fill_color=PLAYER_3_AREAL_COLOR)

    for n, m in PLAYER_4_HOME_TAILES + PLAYER_4_GOAL_AREAL:
        draw_tail(board, n, m, line_color=(0, 0, 0), fill_color=PLAYER_4_AREAL_COLOR)

    for n, m in GLOBS_TIALES:
        put_image_at_taile(board, small_glob, n, m, glob_mask)

    for n, m in STARS_TAILES:
        put_image_at_taile(board, small_star, n, m, star_mask)

    if draw_taile_number:
        draw_taile_indxs(board)

    return board


def get_tailes_player(player_pieces, player):
    tailes_in_use = defaultdict(lambda: 0)

    for i, piece in enumerate(player_pieces):
        if piece == 0:
            taile = HOME_TAILES[player][i]
        else:
            taile = TRACKS[player][piece - 1]

        tailes_in_use[tuple(taile)] += 1

    return list(tailes_in_use.items())


def get_tailes(player_pieces):
    tailes = []
    for player_i, player_pieces in enumerate(player_pieces):
        tailes.append(get_tailes_player(player_pieces, player_i))
    return tailes


def draw_players(board, player_pieces):
    player_tailes = get_tailes(player_pieces)
    for player_i, tailes in enumerate(player_tailes):
        player_color = PLAYER_COLORS[player_i]
        for taile, amount in tailes:
            draw_piece(board, taile[0], taile[1], amount, player_color)


def draw_multi_box(board, top_left_taile, bottom_right_taile, line_color=None, fill_color=None, thickness=2):
    top_left, _, _, _, _ = get_taile_cord(top_left_taile[0], top_left_taile[1])
    _, _, _, bot_right, _ = get_taile_cord(bottom_right_taile[0], bottom_right_taile[1])

    if fill_color is not None:
        cv2.rectangle(board, top_left, bot_right, fill_color, thickness=-1)
    if line_color is not None:
        cv2.rectangle(board, top_left, bot_right, line_color, thickness)


def draw_dice_backgound(board):
    draw_multi_box(board, (3, 17), (4, 18), line_color=(0, 0, 0))
    draw_multi_box(board, (3, 17), (4, 18), fill_color=(255, 255, 255))
    _, _, _, bot_right, center = get_taile_cord(2, 17)
    draw_text(board, f"Dice:", (bot_right[0], center[1]), (0, 0, 0), fontScale=1, thickness=2)


def draw_dice(board, dice, player):
    _, _, _, bot_right, _ = get_taile_cord(3, 17)
    draw_text(board, f"{dice}", bot_right, PLAYER_COLORS[player], fontScale=3, thickness=3)


def draw_move_count_backgound(board):
    draw_multi_box(board, (1, 17), (1, 18), line_color=(0, 0, 0))
    draw_multi_box(board, (1, 17), (1, 18), fill_color=(255, 255, 255))
    _, _, _, bot_right, center = get_taile_cord(0, 17)
    draw_text(board, f"Round:", (bot_right[0], center[1]), (0, 0, 0), fontScale=1, thickness=2)


def draw_move_count(board, count):
    _, _, _, bot_right, center = get_taile_cord(1, 17)
    draw_text(board, f"{count}", (bot_right[0], center[1]), (0, 0, 0), fontScale=1, thickness=2)


def draw_moment(board, moment):
    pieces, dice, players_dice, round_i = moment

    draw_players(board, pieces)
    draw_dice_backgound(board)

    if dice != -1:
        draw_dice(board, dice, players_dice)

    draw_move_count_backgound(board)
    draw_move_count(board, round_i)


def save_hist_video(filename, hist, fps=8, frame_size=None, fourcc=None):
    board = draw_basic_board()

    if frame_size is None:
        save_frame_size = board.shape[:2]
        save_frame_size = (save_frame_size[1], save_frame_size[0])
    else:
        save_frame_size = frame_size

    if fourcc is None:
        file_ext = os.path.split(filename)[-1].split(os.path.extsep)[-1]
        if file_ext == "mp4":
            encoder = cv2.VideoWriter_fourcc(*'MP4V')
        elif file_ext == "avi":
            encoder = cv2.VideoWriter_fourcc(*'MJPG')
        else:
            raise RuntimeError("The video format is not supported. Use a other or add it to the code")
    else:
        encoder = cv2.VideoWriter_fourcc(*fourcc)
    video_out = cv2.VideoWriter(filename, encoder, fps, save_frame_size)

    for i, moment in tqdm(enumerate(hist), total=len(hist)):
        board_take = board.copy()

        draw_moment(board_take, moment)

        if frame_size is not None:
            board_take = cv2.resize(board_take, save_frame_size)

        board_take = cv2.cvtColor(board_take, cv2.COLOR_RGB2BGR)
        video_out.write(board_take)

    video_out.release()


def put_image_at_taile(board, image, n, m, mask=None):
    if mask is None:
        mask = np.full(shape=image.shape[:2], fill_value=True)

    top_left, bot_left, top_right, bot_right, center = get_taile_cord(n, m)

    image_hight, image_width = image.shape[:2]

    board[center[1] - image_hight // 2:center[1] + image_hight // 2,
          center[0] - image_width // 2:center[0] + image_width // 2][mask] = image[mask]


def make_video_from_hist_file(hist_file, video_out, fps=8, frame_size=None, fourcc=None):
    hist = np.load(hist_file, allow_pickle=True)
    save_hist_video(video_out, hist, fps=fps, frame_size=frame_size, fourcc=fourcc)


def make_img_of_board(pieces, dice, players_dice, round_number):
    board = draw_basic_board()
    draw_moment(board, (pieces, dice, players_dice, round_number))
    return board



import numpy as np

TOTAL_NUMBER_OF_TAILES = 60
DICE_MOVE_OUT_OF_HOME = 6
NO_ENEMY = -1

# This roule is that if, there are two pieces on the field, the last one has to return is to start
PLAY_WITH_RULE_A = True

TAILE_FREE = 0
TAILE_HOME = 1
TAILE_START = 2
TAILE_GLOB = 3
TAILE_GOAL_AREAL = 4
TAILE_STAR = 5
TAILE_GOAL = 6
TAILE_ENEMY_1_GLOB = 7
TAILE_ENEMY_2_GLOB = 8
TAILE_ENEMY_3_GLOB = 9
LIST_TAILE_ENEMY_GLOBS = [TAILE_ENEMY_1_GLOB, TAILE_ENEMY_2_GLOB, TAILE_ENEMY_3_GLOB]

NULL_POS = -1
HOME_INDEX = 0
START_INDEX = 1
STAR_INDEXS = [5, 12, 18, 25, 31, 38, 44, 51]

HOME_AREAL_INDEXS = [53, 54, 55, 56, 57, 58]
GOAL_INDEX = 59
GLOB_INDEXS = [9, 22, 35, 48]
ENEMY_1_GLOB_INDX = 14
ENEMY_2_GLOB_INDX = 27
ENEMY_3_GLOB_INDX = 40
STAR_AT_GOAL_AREAL_INDX = STAR_INDEXS[-1]

BORD_TILES = np.full(TOTAL_NUMBER_OF_TAILES, TAILE_FREE)
BORD_TILES[HOME_INDEX] = TAILE_HOME
BORD_TILES[START_INDEX] = TAILE_START
BORD_TILES[STAR_INDEXS] = TAILE_STAR
BORD_TILES[GLOB_INDEXS] = TAILE_GLOB
BORD_TILES[HOME_AREAL_INDEXS] = TAILE_GOAL_AREAL
BORD_TILES[GOAL_INDEX] = TAILE_GOAL
BORD_TILES[ENEMY_1_GLOB_INDX] = TAILE_ENEMY_1_GLOB
BORD_TILES[ENEMY_2_GLOB_INDX] = TAILE_ENEMY_2_GLOB
BORD_TILES[ENEMY_3_GLOB_INDX] = TAILE_ENEMY_3_GLOB

ENEMY_1_INDX_AT_HOME = 40  # HOME_AREAL_INDEXS[0] - 6 - i * 13 # i = 1
ENEMY_2_INDX_AT_HOME = 27  # HOME_AREAL_INDEXS[0] - 6 - i * 13 # i = 2
ENEMY_3_INDX_AT_HOME = 14  # HOME_AREAL_INDEXS[0] - 6 - i * 13 # i = 3


def enemy_pos_at_pos(pos):
    """
    Returns the index's the other players has to be in to be in the same location as the one given in pos

    :param pos: The location to check for
    :type pos: int
    :return enemy_pos: The locations the enemy's pieces has to be at
    :rtype enemy_pos: list of list
    """
    enemy_pos = []

    for enemy_start_pos, enemy_pos_at_start in [[ENEMY_1_GLOB_INDX, ENEMY_1_INDX_AT_HOME],
                                                [ENEMY_2_GLOB_INDX, ENEMY_2_INDX_AT_HOME],
                                                [ENEMY_3_GLOB_INDX, ENEMY_3_INDX_AT_HOME]]:
        post_offset = enemy_start_pos - 1
        pre_offset = enemy_pos_at_start - 1

        if pos == enemy_start_pos:
            pos_enemy = [START_INDEX, HOME_AREAL_INDEXS[0]]
        elif pos < 0:
            pos_enemy = [max(enemy_pos_at_start + pos, -1)]
        elif START_INDEX <= pos < enemy_start_pos:
            pos_enemy = [pos + pre_offset]
        elif pos > HOME_AREAL_INDEXS[0] or pos == HOME_INDEX:
            pos_enemy = [-1]
        else:
            pos_enemy = [pos - post_offset]
        enemy_pos.append(pos_enemy)

    return enemy_pos


def get_enemy_at_pos(pos, enemys):
    """
    Returns the enemy's and the pieces they have at the given location

    :param pos: The location to check for
    :type pos: int
    :param enemys: The locations for the enemy's pieces in a list of 4 lists

    :returns:
    - enemy_at_pos: The enemy's there are at the location
    - enemy_pieces_at_pos: The pieces the enemy's has at the location
    :rtype enemy_at_pos: list
    :rtype enemy_pieces_at_pos: list of list

    """
    # Get the pos the enemy's has to be at to be at the same pos
    other_enemy_pos_at_pos = enemy_pos_at_pos(pos)
    # Check if there is a enemy and how many pieces the enemy has there
    enemy_at_pos = NO_ENEMY
    enemy_pieces_at_pos = []

    for enemy_i, other_enemy_pos in enumerate(other_enemy_pos_at_pos):
        # Check if there already is found a enemy at pos.
        if enemy_at_pos != NO_ENEMY:
            # If there is then stop checking for more (there can only be one)
            break

        for o_pos in other_enemy_pos:
            if o_pos == NULL_POS:
                continue

            for enemy_pice, enemy_pos in enumerate(enemys[enemy_i]):
                if enemy_pos == o_pos:
                    enemy_pieces_at_pos.append(enemy_pice)
                    enemy_at_pos = enemy_i

    return enemy_at_pos, enemy_pieces_at_pos


class Player:
    """
    A class used by the Game class. This class is not needed for normal use
    """

    def __init__(self):
        """
        Makes a player with 4 pieces all at the home locations
        """
        self.pieces = []
        self.number_of_pieces = 4
        self.set_all_pieces_to_home()

    def get_pieces_that_can_move(self, dice):
        """
        Return the pieces that can move with the given dice

        :param dice: The dice the move will be done with
        :type dice: int
        :return: movable_pieces: A list with the pieces that can be moved
        :rtype movable_pieces: list

        """
        movable_pieces = []
        # Go though all the pieces
        for piece_i, piece_place in enumerate(self.pieces):
            # If the piece is a goal then the piece can't move
            if BORD_TILES[piece_place] == TAILE_GOAL:
                continue

            # If the piece is at home and the dice is DICE_MOVE_OUT_OF_HOME then the dice can move out of the home place
            elif BORD_TILES[piece_place] == TAILE_HOME and dice == DICE_MOVE_OUT_OF_HOME:
                movable_pieces.append(piece_i)
            # If the piece is not at home or at the goal it can move
            elif BORD_TILES[piece_place] != TAILE_HOME:
                movable_pieces.append(piece_i)
        return movable_pieces

    def player_winner(self):
        """
        Returns rather the player is a winner or not

        :return: winner: A bool that indicate rather the player is a winner or not
        :rtype winner: bool
        """
        # Go though all the pieces
        for piece_place in self.pieces:
            # If a piece is not at the goal is not the winner
            if BORD_TILES[piece_place] != TAILE_GOAL:
                return False
        # If no piece was not at the goal the player is the winner
        return True

    def set_pieces(self, pieces):
        """
        Sets the players pieces

        :param pieces: The pieces to set the players pieces to
        """
        self.pieces = pieces.copy()

    def get_pieces(self):
        """
        Returns the players pieces

        :return pieces: The players pieces
        :rtype pieces: list
        """
        return self.pieces.copy()

    def move_piece(self, piece, dice, enemys):
        """
        Move the players piece the given dice following the game rules. Returns the new locations of the enemy's pieces

        :param piece: The piece to move
        :type piece: int
        :param dice: The dice to make the move with
        :type dice: int
        :param enemys: The enemy's pieces
        :type enemys: list with 4 lists each with 4 int's
        :return enemys: The new locations of the enemy's pieces
        :rtype enemys: list with 4 lists each with 4 int's

        """
        old_piece_pos = self.pieces[piece]
        new_piece_pos = old_piece_pos + dice

        move_enemy_home_from_poss = []
        do_not_check_rule_a = False
        enemy_at_pos, enemy_pieces_at_pos = get_enemy_at_pos(new_piece_pos, enemys)

        # If the dice is 0 then no movement can be done
        if dice == 0:
            pass

        # At goal
        elif BORD_TILES[old_piece_pos] == TAILE_GOAL:
            # The piece can not move
            pass

        # Goal areal
        elif BORD_TILES[old_piece_pos] == TAILE_GOAL_AREAL:
            if new_piece_pos <= GOAL_INDEX:
                self.pieces[piece] = new_piece_pos
            else:
                overshoot = new_piece_pos - GOAL_INDEX
                new_piece_pos_corrected = old_piece_pos - overshoot
                self.pieces[piece] = new_piece_pos_corrected

        # The Home areal
        elif BORD_TILES[old_piece_pos] == TAILE_HOME:
            if dice == DICE_MOVE_OUT_OF_HOME:
                self.pieces[piece] = START_INDEX

                # Set the enemy there might be at START_INDEX to moved
                do_not_check_rule_a = True
                move_enemy_home_from_poss.append(START_INDEX)

        # Star before the home areal
        elif new_piece_pos == STAR_AT_GOAL_AREAL_INDX:
            self.pieces[piece] = GOAL_INDEX

            # Set the enemy there might be at STAR_AT_GOAL_AREAL_INDX to moved
            move_enemy_home_from_poss.append(new_piece_pos)

        # The other stars
        elif BORD_TILES[new_piece_pos] == TAILE_STAR:
            present_star_staridx = STAR_INDEXS.index(new_piece_pos)
            next_star_staridx = present_star_staridx + 1
            if next_star_staridx >= len(STAR_INDEXS):
                next_star_staridx = 0
            next_star_pos = STAR_INDEXS[next_star_staridx]

            self.pieces[piece] = next_star_pos

            # Set the enemy there might be at first star or the start there will be jump to to be moved
            if enemy_at_pos != NO_ENEMY:
                move_enemy_home_from_poss.append(new_piece_pos)

            next_star_enemy_at_pos, next_star_enemy_pieces_at_pos = get_enemy_at_pos(next_star_pos, enemys)
            if next_star_enemy_at_pos != NO_ENEMY:
                move_enemy_home_from_poss.append(next_star_pos)

        # Globs there are not own by enemy
        elif BORD_TILES[new_piece_pos] == TAILE_GLOB:
            if enemy_at_pos != NO_ENEMY:
                self.pieces[piece] = HOME_INDEX
            else:
                self.pieces[piece] = new_piece_pos

        # Globs there are own by enemy
        elif BORD_TILES[new_piece_pos] in LIST_TAILE_ENEMY_GLOBS:
            # Get the enemy there own the glob
            globs_enemy = LIST_TAILE_ENEMY_GLOBS.index(BORD_TILES[new_piece_pos])
            # Check if there is a enemy at the glob
            if enemy_at_pos != NO_ENEMY:
                # If there is a other enemy then send them home and move there
                if enemy_at_pos != globs_enemy:
                    move_enemy_home_from_poss.append(new_piece_pos)
                    self.pieces[piece] = new_piece_pos
                # If it is the same enemy there is there them move there
                else:
                    self.pieces[piece] = HOME_INDEX
            # If there ant any enemy at the glob then move there
            else:
                self.pieces[piece] = new_piece_pos

        # If it is a TAILE_FREE or if we move from a GLOB/STAR to a not done case
        elif BORD_TILES[old_piece_pos] == TAILE_FREE or \
                BORD_TILES[new_piece_pos] == TAILE_FREE or \
                BORD_TILES[old_piece_pos] == TAILE_GLOB or \
                BORD_TILES[old_piece_pos] == TAILE_STAR:
            if enemy_at_pos != NO_ENEMY:
                move_enemy_home_from_poss.append(new_piece_pos)
            self.pieces[piece] = new_piece_pos

        # If the case was not caught then there is a error
        else:
            print("\nold_piece_pos:", old_piece_pos, "\nnew_piece_pos", new_piece_pos,
                  "\nBORD_TILES[old_piece_pos]:", BORD_TILES[old_piece_pos],
                  "\nBORD_TILES[new_piece_pos]:", BORD_TILES[new_piece_pos], "\ndice:", dice)
            raise RuntimeError("The new_piece_pos case was not handel")

        # Check if there is any enemy there has to be moved
        if len(move_enemy_home_from_poss):
            # Go through the pos where enemy has to be moved from
            for pos in move_enemy_home_from_poss:
                # Get the enemy at the pos
                enemy_at_pos, enemy_pieces_at_pos = get_enemy_at_pos(pos, enemys)
                # Check if there was a enemy at the pos
                if enemy_at_pos != NO_ENEMY:
                    # If there is only one enemy then move the enemy home.
                    if not do_not_check_rule_a and not PLAY_WITH_RULE_A or len(enemy_pieces_at_pos) == 1:
                        for enemy_piece in enemy_pieces_at_pos:
                            enemys[enemy_at_pos][enemy_piece] = HOME_INDEX
                    # If there is more than one then move own piece home
                    else:
                        self.pieces[piece] = HOME_INDEX

        return enemys

    def set_all_pieces_to_home(self):
        """
        Sets all the players pieces to the home index
        """
        self.pieces = []
        for i in range(self.number_of_pieces):
            self.pieces.append(HOME_INDEX)




from .player import Player
from .visualizer import make_img_of_board, save_hist_video
import numpy as np
from tqdm import tqdm


class Game:
    """
    The Game. This class is the only needed class for normal use
    """

    def __init__(self):
        """
        Maked a game with 4 players
        """
        self.players = [Player(), Player(), Player(), Player()]
        self.hist = []
        self.round = 1
        self.current_player = 0
        self.first_winner_was = -1
        self.current_dice = -1
        self.observation_pending = False
        self.current_move_pieces = []
        self.current_enemys = []
        self.current_start_attempts = 0
        self.enemys_order = {
            0: [1, 2, 3],
            1: [2, 3, 0],
            2: [3, 0, 1],
            3: [0, 1, 2]
        }
        self.game_winners = []

    def __dice_generator(self):
        """
        Will update self.current_dice with a new random number on the dice

        """
        self.current_dice = np.random.randint(1, 6 + 1)

    def get_pieces(self, seen_from=None):
        """
        Returns the pieces places on the board

        :param seen_from: indicate which player the pieces and enemy pieces are seen from. If None then the pieces from all 4 player are given and no enemy pieces
        :type seen_from: int
        :returns:
        - pieces: The pieces for alle the players (if seen_from = None) else the pieces for the player given in seen_from
        - enemy_pieces: The pieces of the enemys if a player is given in seen_from
        :rtype pieces: list of 4 int's
        :rtype enemy_pieces: list with 4 lists each with 4 int's

        """
        if seen_from is None:
            pieces = [p.get_pieces() for p in self.players]
            enemy_pieces = None
        else:
            assert 0 <= seen_from <= 3, "The seen_from has to be between 0 and 3. Indicating the player the pieces " \
                                        "are seen from and the nemys are seen from "
            # Get where the players piece are
            pieces = self.players[seen_from].get_pieces()
            # Get where the enemy's piece are
            enemy_pieces = [self.players[e].get_pieces() for e in self.enemys_order[seen_from]]
        return pieces, enemy_pieces

    def __add_to_hist(self):
        """
        Adds the state of the game to the history

        """
        pieces, _ = self.get_pieces()
        self.hist.append([pieces, self.current_dice, self.current_player, self.round])

    def reset(self):
        """
        Resets the game and the game history

        """
        self.players = [Player(), Player(), Player(), Player()]
        self.hist = []
        self.round = 1
        self.current_player = 0
        self.first_winner_was = -1
        self.current_dice = -1
        self.observation_pending = False
        self.current_move_pieces = []
        self.current_enemys = []
        self.current_start_attempts = 0
        self.game_winners = []

    def __gen_observation(self, player_idx, roll_dice = True):
        if roll_dice:
            # Roll the dice
            self.__dice_generator()
        dice = self.current_dice

        player = self.players[player_idx]
        # Get the pieces there can be moved with the current dice
        move_pieces = player.get_pieces_that_can_move(dice)
        self.current_move_pieces = move_pieces

        # Get where the players piece are and the enemy's piece are
        player_pieces, enemy_pieces = self.get_pieces(player_idx)
        self.current_enemys = enemy_pieces
        # Check if the player is a winner
        player_is_a_winner = player.player_winner()
        # Check if there is a winner
        there_is_a_winner = any([p.player_winner() for p in self.players])

        return dice, move_pieces, player_pieces, enemy_pieces, player_is_a_winner, there_is_a_winner

    def __set_enemy_pieces(self, player_idx, enemy_pieces):
        """
        Will set the enemy pieces to the pieces given in enemy_pieces

        :param player_idx: The player the enemys are seen from
        :type player_idx: int
        :param enemy_pieces: The pieces to update
        :type enemy_pieces: list with 4 lists each with 4 int's
        """
        # Go throng the enemy's and set the changes in there piece
        for e_i, e in enumerate(self.enemys_order[player_idx]):
            self.players[e].set_pieces(enemy_pieces[e_i])

    def get_observation(self):
        """
        Return the state/observation of the game and which players turn it is
        A given observation has to be answered before a new one can be given.

        :returns:
        - obs: The observation taken of the state of the game seen from the player given in the return current_player (dice, move_pieces, player_pieces, enemy_pieces, player_is_a_winner, there_is_a_winner)
        - current_player: Which players turn it is
        :rtype obs: (int, list with upto 4 int's, list with 4 int's, list of 4 lists with 4 int's, bool, bool)
        :rtype current_player: int

        """
        # Check if there is a observation pending
        if self.observation_pending:
            raise RuntimeError("There is already a pending observation. "
                               "The pending observation has to be answered first")
        # Set pending observation to true
        self.observation_pending = True
        # Get the current en environment
        obs = self.__gen_observation(self.current_player, roll_dice=True)

        # Add the bord and dice before the move to the history
        self.__add_to_hist()
        return obs, self.current_player

    def __count_player(self):
        """
        Updates the current player and the round if the current player was the last player

        """
        # Count up the player
        self.current_player += 1
        # If the count is over 3 then reset to player 0 and count up the round
        if self.current_player > 3:
            self.current_player = 0
            self.round += 1

    def answer_observation(self, piece_to_move):
        """
        Answers a observation. A observation has to be given before a answer can be given.

        :param piece_to_move: Which piece to move. If there was no pieces there cloud be moved the parameter is ignored
        :type piece_to_move: int
        :return obs: Who the game was after the given move was done. obs is: (dice, move_pieces, player_pieces, enemy_pieces, player_is_a_winner, there_is_a_winner)
        :rtype obs: (int, list with upto 4 int's, list with 4 int's, list of 4 lists with 4 int's, bool, bool)
        """
        # Check if there is a observation pending
        if not self.observation_pending:
            raise RuntimeError("There is no pending observation. "
                               "There has to be a pending observation has to be answered first")
        # Check if the given piece_to_move is among the current_move_pieces
        if len(self.current_move_pieces) and piece_to_move not in self.current_move_pieces:
            raise RuntimeError("The piece given has to be among the given move_pieces")
        # If it is then move the piece
        elif len(self.current_move_pieces):
            new_enemys = self.players[self.current_player].move_piece(piece_to_move,
                                                                      self.current_dice, self.current_enemys)
            self.__set_enemy_pieces(self.current_player, new_enemys)
        # If there was no pieces there could be moved then nothing can be done
        else:
            pass # This line is present for readability

        # Check if the player now is the winner
        player_is_a_winner = self.players[self.current_player].player_winner()
        if player_is_a_winner:
            # Check if player is the first winner
            if self.first_winner_was == -1:
                self.first_winner_was = self.current_player
            # Check if player has been added to game_winners
            if self.current_player not in self.game_winners:
                self.game_winners.append(self.current_player)

        # Add the bord after the move to the history
        self.__add_to_hist()

        next_player = True
        # In the first round the players has 3 attempts to get a piece out of home
        if self.round == 1 and \
                all(p_piece == 0 for p_piece in self.players[self.current_player].get_pieces()) and \
                self.current_start_attempts < 3:
            self.current_start_attempts += 1
            next_player = False
        else:
            self.current_start_attempts = 0
        # If it is not in the first round a dice on 6 will give a extra move
        if self.round != 1 and self.current_dice == 6:
            next_player = False
        # If it is the next players turn then change current_player
        if next_player:
            self.__count_player()

        # Set the observation pending to false as the last given observation was handled
        self.observation_pending = False

        # Get the environment after the move
        after_obs = self.__gen_observation(self.current_player, roll_dice=False)

        return after_obs

    def get_winner_of_game(self):
        """
        Returns the winner of the game

        :return winner: If there has been a winner the winner is return if not -1 is returned
        :rtype winner: bool
        """
        return self.first_winner_was

    def get_winners_of_game(self):
        """
        Returns the winners of the game

        :return gameWinners: A list of the winners of the game in the order they got all piece in goal
        :rtype gameWinners: list with upto 4 int's
        """
        return self.game_winners

    def all_players_finish(self):
        """
        Returns rather all players has finish

        :return allFinish: Bool rather all players has finish the game
        :rtype allFinish: bool
        """
        return len(self.game_winners) == len(self.players)

    def get_hist(self):
        """
        Returns the history there has been recorded during the game. This history can be used to make
        a video of the game. The history will have been extended when a observation was given and when a
        answer to a observation was given.

        :return hist: list of [pieces, current_dice, first_winner_was, current_player, round]
        :rtype hist: [list with 4 lists with 4 int's, int, bool, int, int]
        """
        return self.hist

    def get_piece_hist(self, mode=0):
        """
        Will return the how the pieces was recorded during the game.

        :param mode: 0: All recorded pieces is returnt. 1: Only if a change is done there will be a new set of pieces. 2: Only unique set of pieces (order is preserved)
        :type mode: int
        :return piece_hist: List of sets of pieces [player 1, player 2, player 3, player 4]
        :rtype piece_hist: list of 4 lists with 4 int's
        """
        piece_hist = [self.hist[0][0]]
        for h in self.hist[1:]:
            pieces = h[0]

            add_to_hist = False
            if mode == 0:
                add_to_hist = True
            if mode == 1:
                if pieces != piece_hist[-1]:
                    add_to_hist = True
            if mode == 2:
                if pieces not in piece_hist:
                    add_to_hist = True

            if add_to_hist:
                piece_hist.append(pieces)
        return piece_hist

    def render_environment(self):
        """
        Will render the last record in the history

        :return board_img: A image of the board
        :rtype board_img: ndarray, RGB colorspace
        """
        board_img = make_img_of_board(*self.hist[-1])
        return board_img

    def save_hist(self, file_name):
        """
        Saves the history of the game as a npy file

        :param file_name: The file name to save under. Has to have the .npy (numpy file) extension
        :type file_name: str

        """
        file_ext = file_name.split(".")[-1]
        assert file_ext == "npy", "The file extension has to be npy (numpy file)"
        np.save(file_name, self.hist)

    def save_hist_video(self, video_out, fps=8, frame_size=None, fourcc=None):
        """
        Saves a video of the game history

        :param video_out: The file name to save under
        :type video_out: str
        :param fps: Frames per second
        :type fps: float
        :param frame_size: The frame size to save in (width, height). If None is given the full board size is used
        :type frame_size: tuple
        :param fourcc: FourCC code to be used. If None is given the FourCC code will be tried to create fro the file extension (works on .mp4 and .avi)
        :type fourcc: str

        """
        save_hist_video(video_out, self.hist, fps=fps, frame_size=frame_size, fourcc=fourcc)




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ludopy",
    version="1.1.10",
    author="Simon L. B. Sørensen",
    author_email="simonlyckbjaert@hotmail.com",
    description="LUDO python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SimonLBSoerensen/LUDOpy",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords="LUDO, game",
    python_requires='>=3.6',
    license='MIT',
    install_requires=["numpy", "opencv-python>=3.1", "tqdm"],
    include_package_data=True,
)

#python setup.py sdist bdist_wheel
#python -m twine upload dist/*
#python -m twine upload --repository-url https://test.pypi.org/legacy/ dist/*